import metallurgy as mg


def test_melting_temperature():
    assert mg.deviation("Cu", "atomic_number") == 0

    assert (
        mg.deviation({"Cu": 0.5, "Zr": 0.5}, "melting_temperature") == 385.115
    )



import metallurgy as mg


def test_alloy_creation():

    alloy_from_string = mg.Alloy("Fe20Ag80")

    alloy_from_dict = mg.Alloy({"Fe": 20, "Ag": 80})

    assert alloy_from_string.__dict__ == alloy_from_dict.__dict__


def test_alloy_to_string():
    alloy = mg.Alloy({"Fe": 20, "Ag": 80})
    assert alloy.to_string() == "Ag80Fe20"



import numpy as np
from ovito.io import import_file, export_file
from ovito.modifiers import CoordinationAnalysisModifier


def rdf(structure_path):

    pipeline = import_file(structure_path)
    pipeline.modifiers.append(
        CoordinationAnalysisModifier(cutoff=10, number_of_bins=150)
    )

    data = pipeline.compute()

    return data.tables["coordination-rdf"]["g(r)"]


def smoothness(structure_path, filter=False):

    RDF = rdf(structure_path)

    RDF_diffs = []
    for i in range(1, len(RDF)):
        RDF_diffs.append(RDF[i] - RDF[i - 1])

    smoothness = np.abs(np.mean(RDF_diffs)) / np.std(RDF_diffs)

    return smoothness



from elementy.periodictable import PeriodicTable

from .linear_mixture import linear_mixture
from .deviation import deviation
from .alloy import Alloy
from . import features
from . import density
from . import valence
from . import entropy
from . import enthalpy
from . import radii
from . import viscosity
from . import price
from . import generate
from . import ratios
from .calculate import calculate
from . import plots
from . import analyse

# from . import rdf

periodic_table = PeriodicTable()
model = None


def set_model(model_in):
    global model
    model = model_in


def get_model():
    return model


__all__ = [
    "periodic_table",
    "Alloy",
    "linear_mixture",
    "deviation",
    "density",
    "valence",
    "entropy",
    "enthalpy",
    "radii",
    "price",
    "viscosity",
    "ratios",
    # "rdf",
    "plots",
    "generate",
    "calculate",
    "features",
    "analyse",
    "model",
    "set_model",
    "get_model",
]



from collections.abc import Iterable

from . import linear_mixture
from .alloy import Alloy


def shell_valence_electron_concentration_ratio(
    alloy, period=None, valence_electrons=None
):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [shell_valence_electron_concentration_ratio(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    if period is None:
        period = linear_mixture(alloy, "period")
    if valence_electrons is None:
        valence_electrons = linear_mixture(alloy, "valence_electrons")

    return period / valence_electrons


def shell_mendeleev_number_ratio(alloy, period=None, mendeleev=None):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [shell_mendeleev_number_ratio(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    if period is None:
        period = linear_mixture(alloy, "period")
    if mendeleev is None:
        mendeleev = linear_mixture(alloy, "mendeleev_universal_sequence")

    if period is None or mendeleev is None:
        return None
    return period / mendeleev



from collections.abc import Iterable

import metallurgy as mg
from . import linear_mixture
from .alloy import Alloy


def valence_proportion(alloy, orbital):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [valence_proportion(a, orbital) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    total_valence = linear_mixture(alloy, "valence_electrons")

    orbitalCount = {}
    for element in alloy.elements:
        orbitalCount[element] = 0

        valence_electrons = mg.periodic_table.elements[element][
            "valence_electrons"
        ]
        orbitals = mg.periodic_table.elements[element]["orbitals"]

        i = 0
        electrons = 0
        while electrons < valence_electrons:
            electrons += orbitals[-1 - i]["electrons"]
            if orbitals[-1 - i]["orbital"][-1] == orbital:
                orbitalCount[element] += orbitals[-1 - i]["electrons"]
            i += 1

    if total_valence > 0:
        total = 0
        for element in alloy.elements:
            total += orbitalCount[element] * alloy.composition[element]
        return total / total_valence

    else:
        return 0


def s_valence(alloy):
    return valence_proportion(alloy, "s")


def p_valence(alloy):
    return valence_proportion(alloy, "p")


def d_valence(alloy):
    return valence_proportion(alloy, "d")


def f_valence(alloy):
    return valence_proportion(alloy, "f")



import re

import numpy as np

from . import generate


def find_max(elements, feature_name):
    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    system = generate.system(elements, feature_name=feature_name)
    if system is not None:

        alloys, percentages, values = system

        max_index = np.argmax(values)

        return alloys[max_index], values[max_index]


def find_min(elements, feature_name):
    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    system = generate.system(elements, feature_name=feature_name)
    if system is not None:

        alloys, percentages, values = system

        min_index = np.argmin(values)

        return alloys[min_index], values[min_index]



import pytest
import metallurgy as mg


def test_mixing_enthalpy():
    assert mg.enthalpy.mixing_enthalpy("Cu") == 0
    assert mg.enthalpy.mixing_enthalpy("Cu50Zr50") == pytest.approx(
        -22.598395865085948
    )


def test_mixing_Gibbs_free_energy():
    assert mg.enthalpy.mixing_Gibbs_free_energy("Cu") == 0
    assert mg.enthalpy.mixing_Gibbs_free_energy("CuZr") == pytest.approx(
        -35625.153905259554
    )


def test_mismatch_PHS():
    assert mg.enthalpy.mismatch_PHS("Fe") == 0
    assert mg.enthalpy.mismatch_PHS("CuZr") == pytest.approx(
        -4.661624378630986
    )


def test_mixing_PHS():
    assert mg.enthalpy.mixing_PHS("Fe") == 0
    assert mg.enthalpy.mixing_PHS("CuZr") == pytest.approx(-20.325638757692836)


def test_mixing_PHSS():
    assert mg.enthalpy.mixing_PHSS("Fe") == 0
    assert mg.enthalpy.mixing_PHSS("CuZr") == pytest.approx(-4.192797298966486)


def test_thermodynamic_factor():
    assert mg.enthalpy.thermodynamic_factor("Fe") == 0
    assert mg.enthalpy.thermodynamic_factor("CuZr") == pytest.approx(
        0.06936776640549777
    )


def test_Gamma():
    assert mg.enthalpy.Gamma("Cu", "Zr") == pytest.approx(-13.605582368214959)


def test_calculate_QPR():
    assert mg.enthalpy.calculate_QPR("Cu", "Fe") == (132.54, 14.1, 0)
    assert mg.enthalpy.calculate_QPR("Fe", "C") == (115.62, 12.3, 2.1)
    assert mg.enthalpy.calculate_QPR("B", "C") == (99.64, 10.6, 0)


def test_wigner_seitz_discontinuity():
    assert mg.enthalpy.wigner_seitz_electron_density_discontinuity_delta(
        "Cu", "Zr"
    ) == pytest.approx(0.06107640860538699)


def test_topological_enthalpy():
    assert mg.enthalpy.calculate_topological_enthalpy(
        {"Cu": 0.5, "Zr": 0.5}
    ) == pytest.approx(13.629999999999999)



import pytest
import metallurgy as mg


def test_mixing_entropy():
    assert mg.entropy.mixing_entropy("Cu") == 0
    assert mg.entropy.mixing_entropy("Cu50Zr50") == pytest.approx(
        0.8994283877067366
    )
    assert mg.entropy.mixing_entropy(["Fe", "Cu50Zr50"]) == [
        0,
        pytest.approx(0.8994283877067366),
    ]


def test_mismatch_entropy():
    assert mg.entropy.mismatch_entropy("Fe") == 0
    assert mg.entropy.mismatch_entropy("Cu50Zr50") == pytest.approx(
        0.2062812071467913
    )
    assert mg.entropy.mismatch_entropy(["Fe", "Cu50Zr50"]) == [
        0,
        pytest.approx(0.2062812071467913),
    ]


def test_ideal_entropy():
    assert mg.entropy.ideal_entropy("Xe") == 0
    assert mg.entropy.ideal_entropy("XeAr") == pytest.approx(
        0.6931471805599453
    )
    assert mg.entropy.ideal_entropy("GdFeNi") == pytest.approx(
        1.0986112890009436
    )
    assert mg.entropy.ideal_entropy(["Xe", "XeAr", "GdFeNi"]) == [
        0,
        pytest.approx(0.6931471805599453),
        pytest.approx(1.0986112890009436),
    ]


def test_ideal_entropy_xia():
    assert mg.entropy.ideal_entropy_xia("Xe") == 0
    assert mg.entropy.ideal_entropy_xia("XeAr") == pytest.approx(
        0.7343892084076673
    )
    assert mg.entropy.ideal_entropy_xia("XeArNi") == pytest.approx(
        1.4161052076473997
    )
    assert mg.entropy.ideal_entropy_xia(["Xe", "XeAr", "XeArNi"]) == [
        0,
        pytest.approx(0.7343892084076673),
        pytest.approx(1.4161052076473997),
    ]



from collections.abc import Iterable

import numpy as np

import metallurgy as mg
from .alloy import Alloy
from . import linear_mixture
from . import constants
from . import entropy


def Gamma(elementA, elementB):
    Q, P, R = calculate_QPR(elementA, elementB)
    if R is not None:
        return (
            calculate_electronegativity_enthalpy_component(
                elementA, elementB, P
            )
            + calculate_WS_enthalpy_component(elementA, elementB, Q)
            - R
        )
    else:
        return None


def calculate_QPR(elementA, elementB):

    seriesA = mg.periodic_table.elements[elementA]["series"]
    if elementA == "Ca" or elementA == "Sr" or elementA == "Ba":
        seriesA = "non_transition_metal"

    seriesB = mg.periodic_table.elements[elementB]["series"]
    if elementB == "Ca" or elementB == "Sr" or elementB == "Ba":
        seriesB = "non_transition_metal"

    if seriesA == "transition_metal" and seriesB == "transition_metal":
        P = 14.1
        R = 0
    elif seriesA != "transition_metal" and seriesB != "transition_metal":
        P = 10.6
        R = 0
    else:
        P = 12.3

        if (
            mg.periodic_table.elements[elementA]["miedema_R"] is not None
            and mg.periodic_table.elements[elementB]["miedema_R"] is not None
        ):
            R = (
                mg.periodic_table.elements[elementA]["miedema_R"]
                * mg.periodic_table.elements[elementB]["miedema_R"]
            )
        else:
            R = None

    Q = P * 9.4

    return Q, P, R


def calculate_electronegativity_enthalpy_component(elementA, elementB, P):
    electronegativityDiff = (
        mg.periodic_table.elements[elementA]["electronegativity_miedema"]
        - mg.periodic_table.elements[elementB]["electronegativity_miedema"]
    )
    return -P * (electronegativityDiff**2)


def calculate_WS_enthalpy_component(elementA, elementB, Q):
    return Q * (
        wigner_seitz_electron_density_discontinuity_delta(elementA, elementB)
        ** 2
    )


def wigner_seitz_electron_density_discontinuity_delta(elementA, elementB):
    densityA = mg.periodic_table.elements[elementA][
        "wigner_seitz_electron_density"
    ]
    densityB = mg.periodic_table.elements[elementB][
        "wigner_seitz_electron_density"
    ]
    return (densityA ** (1.0 / 3.0)) - (densityB ** (1.0 / 3.0))


def calculate_surface_concentration(elements, volumes, composition):

    reduced_vol_A = composition[elements[0]] * (volumes[0] ** (2.0 / 3.0))
    reduced_vol_B = composition[elements[1]] * (volumes[1] ** (2.0 / 3.0))

    return reduced_vol_A / (reduced_vol_A + reduced_vol_B)


def calculate_corrected_volume(elementA, elementB, Cs_A):

    pureV = mg.periodic_table.elements[elementA]["volume_miedema"]

    electronegativityDiff = (
        mg.periodic_table.elements[elementA]["electronegativity_miedema"]
        - mg.periodic_table.elements[elementB]["electronegativity_miedema"]
    )

    a = None
    if elementA in ["Ca", "Sr", "Ba"]:
        a = 0.04
    elif elementA in ["Ru", "Rh", "Pd", "Os", "Ir", "Pt", "Au"]:
        a = 0.07

    if a is None:
        if mg.periodic_table.elements[elementA]["series"] == "alkaliMetal":
            a = 0.14
        elif mg.periodic_table.elements[elementA]["valence_electrons"] == 2:
            a = 0.1
        elif mg.periodic_table.elements[elementA]["valence_electrons"] == 3:
            a = 0.07
        else:
            a = 0.04

    f_AB = 1 - Cs_A

    correctedV = (pureV ** (2.0 / 3.0)) * (
        1 + a * f_AB * electronegativityDiff
    )

    return correctedV


def calculate_interface_enthalpy(elementA, elementB, volumeA):

    densityA = mg.periodic_table.elements[elementA][
        "wigner_seitz_electron_density"
    ]
    densityB = mg.periodic_table.elements[elementB][
        "wigner_seitz_electron_density"
    ]

    gamma = Gamma(elementA, elementB)
    if gamma is not None:
        return (
            2
            * volumeA
            * gamma
            / (densityA ** (-1.0 / 3.0) + densityB ** (-1.0 / 3.0))
        )


def calculate_topological_enthalpy(composition):
    topological_enthalpy = 0
    for element in composition:
        topological_enthalpy += (
            mg.periodic_table.elements[element]["fusion_enthalpy"]
            * composition[element]
        )

    return topological_enthalpy


def mixing_enthalpy(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_enthalpy(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    if len(alloy.elements) > 1:

        elementPairs = [
            (a, b)
            for idx, a in enumerate(alloy.elements)
            for b in alloy.elements[idx + 1 :]
        ]

        total_mixing_enthalpy = 0
        for pair in elementPairs:
            tmpComposition = {}
            subComposition = 0
            for element in pair:
                subComposition += alloy.composition[element]
            for element in pair:
                tmpComposition[element] = (
                    alloy.composition[element] / subComposition
                )

            Cs_A = None
            V_A_alloy = mg.periodic_table.elements[pair[0]]["volume_miedema"]
            V_B_alloy = mg.periodic_table.elements[pair[1]]["volume_miedema"]

            if V_A_alloy is None or V_B_alloy is None:
                return None

            for _ in range(10):

                Cs_A = calculate_surface_concentration(
                    pair, [V_A_alloy, V_B_alloy], tmpComposition
                )

                V_A_alloy = calculate_corrected_volume(pair[0], pair[1], Cs_A)

                V_B_alloy = calculate_corrected_volume(
                    pair[1], pair[0], 1 - Cs_A
                )

            interface_AB = calculate_interface_enthalpy(
                pair[0], pair[1], V_A_alloy
            )
            interface_BA = calculate_interface_enthalpy(
                pair[1], pair[0], V_B_alloy
            )

            if interface_AB is not None and interface_BA is not None:

                chemical_enthalpy = (
                    alloy.composition[pair[0]]
                    * alloy.composition[pair[1]]
                    * ((1 - Cs_A) * interface_AB + Cs_A * interface_BA)
                )
                total_mixing_enthalpy += chemical_enthalpy

            else:
                return None

    else:
        total_mixing_enthalpy = 0.0

    return total_mixing_enthalpy


def mixing_Gibbs_free_energy(alloy):

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_Gibbs_free_energy(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    H = mixing_enthalpy(alloy)
    Tm = linear_mixture(alloy, "melting_temperature")
    S = entropy.mixing_entropy(alloy)

    if H is None or Tm is None or S is None:
        return None
    return (H * 1e3) - Tm * S * constants.idealGasConstant


def mismatch_PHS(alloy):

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mismatch_PHS(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    H = mixing_enthalpy(alloy)
    S = entropy.mismatch_entropy(alloy)

    if H is not None and S is not None:
        return H * S
    else:
        return None


def mixing_PHS(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_PHS(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    H = mixing_enthalpy(alloy)
    S = entropy.mixing_entropy(alloy)

    if H is not None and S is not None:
        return H * S
    else:
        return None


def mixing_PHSS(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_PHSS(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    H = mixing_enthalpy(alloy)
    Smix = entropy.mixing_entropy(alloy)
    Smismatch = entropy.mismatch_entropy(alloy)

    if H is None or Smix is None or Smismatch is None:
        return None
    else:
        return H * Smix * Smismatch


def thermodynamic_factor(alloy):

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [thermodynamic_factor(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    Tm = linear_mixture(alloy, "melting_temperature")
    H = mixing_enthalpy(alloy)
    S = entropy.mixing_entropy(alloy)

    if Tm is None or H is None or S is None:
        return None

    return (Tm * S) / (np.abs(H * 1e3) + 1e-10)



import numbers
from collections.abc import Iterable

import numpy as np
import metallurgy as mg

from .alloy import Alloy


def deviation(alloy, feature_name):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [deviation(a, feature_name) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    if len(alloy.elements) > 1:

        if isinstance(
            mg.periodic_table.elements[alloy.elements[0]][feature_name],
            (int, float, list),
        ):

            mean = 0
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][feature_name]
                if value is None:
                    return None

                if isinstance(value, list):
                    value = value[0]
                if not isinstance(value, numbers.Number):
                    return None

                mean += alloy.composition[element] * value

            total_deviation = 0
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][feature_name]

                if isinstance(value, list):
                    value = value[0]
                if not isinstance(value, numbers.Number):
                    return None

                total_deviation += alloy.composition[element] * (
                    (value - mean) ** 2
                )

            return total_deviation**0.5

        else:
            values = {}
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][feature_name]
                if value is None:
                    return None

                if value not in values:
                    values[value] = 0
                values[value] += alloy.composition[element]

            if len(values) > 1:
                shannonEntropy = 0
                for value in values:
                    shannonEntropy -= values[value] * np.log(values[value])
                return shannonEntropy
            else:
                return 0

    else:
        return 0



import metallurgy as mg


def test_melting_temperature():
    assert (
        mg.linear_mixture({"Cu": 0.5, "Zr": 0.5}, "melting_temperature")
        == 1742.885
    )

    assert mg.linear_mixture(
        [{"Cu": 0.5, "Zr": 0.5}, {"Cu": 0.25, "Zr": 0.75}],
        "melting_temperature",
    ) == [1742.885, 1935.4425]



import re
from typing import Union, Optional

import numpy as np
import elementy

from .alloy import Alloy
from .calculate import calculate


def random_alloy(
    min_elements: int = 1,
    max_elements: int = 10,
    percentage_constraints={},
    allowed_elements=[e for e in elementy.PeriodicTable().elements],
):

    if isinstance(percentage_constraints, (list)):
        parsed_percentage_constraints = {}
        for element in percentage_constraints:
            parsed_percentage_constraints[element] = {}
        percentage_constraints = parsed_percentage_constraints

    if min_elements < max_elements:
        num_extra_elements = np.random.randint(min_elements, max_elements)
    else:
        num_extra_elements = max_elements
    num_extra_elements -= len(percentage_constraints)

    if num_extra_elements > 0:
        other_elements = allowed_elements[:]
        for element in percentage_constraints:
            if element in other_elements:
                other_elements.remove(element)

        elements = list(percentage_constraints.keys()) + list(
            np.random.choice(
                other_elements,
                min(num_extra_elements, len(other_elements)),
                replace=False,
            )
        )

    else:
        elements = list(
            np.random.choice(
                list(percentage_constraints.keys()),
                num_extra_elements + len(percentage_constraints.keys()),
                replace=False,
            )
        )

    percentages = list(np.random.dirichlet(np.ones(len(elements)), size=1)[0])

    composition = {}
    for j in range(len(elements)):
        composition[elements[j]] = percentages[j]

    alloy = Alloy(
        composition,
        constraints={
            "percentages": percentage_constraints,
            "min_elements": min_elements,
            "max_elements": max_elements,
        },
    )

    return alloy


def random_alloys(
    num_alloys: int,
    min_elements: int = 1,
    max_elements: int = 10,
    percentage_constraints: dict = {},
    allowed_elements: list = [e for e in elementy.PeriodicTable().elements],
):

    return [
        random_alloy(
            min_elements,
            max_elements,
            percentage_constraints,
            allowed_elements,
        )
        for _ in range(num_alloys)
    ]


def mixture(alloys, weights: Optional[list] = None):

    shared_composition_space = []
    for alloy in alloys:
        for element in alloy.elements:
            if element not in shared_composition_space:
                shared_composition_space.append(element)

    constraints = None
    for alloy in alloys:
        if alloy.constraints is not None:
            if constraints is None:
                constraints = {
                    "percentages": {},
                    "min_elements": 1,
                    "max_elements": 1,
                }

            if "min_elements" in alloy.constraints:
                constraints["min_elements"] = max(
                    constraints["min_elements"],
                    alloy.constraints["min_elements"],
                )
            if "max_elements" in alloy.constraints:
                constraints["max_elements"] = max(
                    constraints["max_elements"],
                    alloy.constraints["max_elements"],
                )

            if "percentages" in alloy.constraints:
                for element in alloy.constraints["percentages"]:
                    if element not in constraints:
                        constraints["percentages"][element] = {}

                    if "min" in alloy.constraints["percentages"][element]:
                        if "min" not in constraints["percentages"][element]:
                            constraints["percentages"][element][
                                "min"
                            ] = alloy.constraints["percentages"][element][
                                "min"
                            ]
                        else:
                            constraints["percentages"][element]["min"] = min(
                                constraints["percentages"][element]["min"],
                                alloy.constraints["percentages"][element][
                                    "min"
                                ],
                            )

                    if "max" in alloy.constraints["percentages"][element]:
                        if "max" not in constraints["percentages"][element]:
                            constraints["percentages"][element][
                                "max"
                            ] = alloy.constraints["percentages"][element][
                                "max"
                            ]
                        else:
                            constraints["percentages"][element]["max"] = max(
                                constraints["percentages"][element]["max"],
                                alloy.constraints["percentages"][element][
                                    "max"
                                ],
                            )

    if weights is None:
        weights = [1.0] * len(alloys)

    mixed_composition = {}
    for element in shared_composition_space:
        mixed_composition[element] = 0
        for i in range(len(alloys)):
            if element in alloys[i].elements:
                mixed_composition[element] += (
                    alloys[i].composition[element] * weights[i]
                )

    return Alloy(mixed_composition, constraints=constraints)


def system(
    elements: Union[list, str],
    step: Union[int, float] = 1,
    min_percent: Union[int, float] = 0,
    max_percent: Union[int, float] = 100,
    feature_name: Optional[str] = None,
    quaternary: Optional[dict] = None,
):

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    if len(elements) == 2:
        return binary(elements, step, feature_name)
    elif len(elements) == 3:
        return ternary(
            elements, step, min_percent, max_percent, feature_name, quaternary
        )


def binary(
    elements: Union[list, str],
    step: Union[int, float] = 0.5,
    feature_name: Optional[str] = None,
):

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    x = 100
    alloys = []
    percentages = []
    while x >= 0:
        alloys.append(
            Alloy(
                elements[0] + str(x) + elements[1] + str(100 - x),
                rescale=False,
            )
        )
        percentages.append(x)
        x -= step

    if feature_name is not None:
        values = calculate(alloys, feature_name)
        return alloys, percentages, values

    return alloys, percentages


def ternary(
    elements: Union[list, str],
    step: Union[int, float] = 1,
    min_percent: Union[int, float] = 0,
    max_percent: Union[int, float] = 100,
    feature_name: Optional[str] = None,
    quaternary: Optional[dict] = None,
):

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    alloys = []
    percentages = []

    tmp_percentages = [max_percent + step] * 3
    while tmp_percentages[0] >= min_percent + step:
        tmp_percentages[0] -= step
        tmp_percentages[1] = max_percent + step
        tmp_percentages[2] = max_percent + step

        while tmp_percentages[1] >= min_percent + step:
            tmp_percentages[1] -= step
            tmp_percentages[2] = max_percent + step

            while tmp_percentages[2] >= min_percent + step:
                tmp_percentages[2] -= step

                if sum(tmp_percentages) == 100:

                    composition_str = ""
                    for i in range(len(elements)):
                        if tmp_percentages[i] > 0:
                            composition_str += elements[i] + str(
                                tmp_percentages[i]
                            )

                    if quaternary is not None:
                        composition_str = (
                            "("
                            + composition_str
                            + ")"
                            + str(100 - quaternary[1])
                            + quaternary[0]
                            + str(quaternary[1])
                        )

                    alloy = Alloy(composition_str, rescale=False)

                    alloys.append(alloy)
                    percentages.append(list(alloy.composition.values()))

    if feature_name is not None:
        values = calculate(alloys, feature_name)
        return alloys, percentages, values

    return alloys, percentages



from collections.abc import Iterable

import numpy as np

import metallurgy as mg
from . import enthalpy
from . import linear_mixture
from . import constants
from .alloy import Alloy


def viscosity(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [viscosity(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    const = 3.077e-3
    elementalViscosity = {}
    for element in alloy.elements:
        mass = mg.periodic_table.elements[element]["mass"]
        Tm = mg.periodic_table.elements[element]["melting_temperature"]
        molar_volume = mg.periodic_table.elements[element]["molar_volume"]
        if mass is None or Tm is None or molar_volume is None:
            return None

        elementalViscosity[element] = (
            const * np.sqrt((mass / 1000) * Tm) / (molar_volume * 1.0e-6)
        )

    sum_aG = 0
    for element in alloy.elements:
        mass = mg.periodic_table.elements[element]["mass"]
        density = mg.periodic_table.elements[element]["density"]
        Tm = mg.periodic_table.elements[element]["melting_temperature"]
        if mass is None or density is None or Tm is None:
            return None

        sum_aG += (
            Tm
            * alloy.composition[element]
            * np.log(
                (elementalViscosity[element] * (mass / 1000))
                / (
                    constants.plankConstant
                    * constants.avogadroNumber
                    * (density)
                    * 1000
                )
            )
        )

    sum_aG *= constants.idealGasConstant

    averageMolarVolume = 0
    for element in alloy.elements:
        averageMolarVolume += alloy.composition[element] * (
            mg.periodic_table.elements[element]["molar_volume"] * 1.0e-6
        )

    H = enthalpy.mixing_enthalpy(alloy)
    if H is None:
        return None

    return (
        (constants.plankConstant * constants.avogadroNumber)
        / (averageMolarVolume)
    ) * np.exp(
        (sum_aG - 0.155 * H)
        / (
            constants.idealGasConstant
            * linear_mixture(alloy, "melting_temperature")
        )
    )



from collections.abc import Iterable

import metallurgy as mg
from .alloy import Alloy


def theoretical_density(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [theoretical_density(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    massFractions = {}
    masses = {}
    totalMass = 0
    for element in alloy.elements:
        masses[element] = mg.periodic_table.elements[element]["mass"]
        totalMass += masses[element] * alloy.composition[element]

    for element in alloy.elements:
        massFractions[element] = (
            alloy.composition[element] * masses[element] / totalMass
        )

    total = 0
    for element in alloy.elements:
        total += (
            massFractions[element]
            / mg.periodic_table.elements[element]["density"]
        )

    return 1 / total



from collections.abc import Iterable

import numpy as np

import metallurgy as mg
from .alloy import Alloy


def radius_gamma(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [radius_gamma(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    maxR = 0
    minR = 1000

    meanR = 0
    for element in alloy.elements:
        meanR += (
            alloy.composition[element]
            * mg.periodic_table.elements[element]["radius"]
        )

    for element in alloy.elements:
        r = mg.periodic_table.elements[element]["radius"]
        if r > maxR:
            maxR = r
        if r < minR:
            minR = r

    rMinAvSq = (minR + meanR) ** 2
    rMaxAvSq = (maxR + meanR) ** 2
    rAvSq = meanR**2

    numerator = 1.0 - np.sqrt((rMinAvSq - rAvSq) / (rMinAvSq))
    denominator = 1.0 - np.sqrt((rMaxAvSq - rAvSq) / (rMaxAvSq))

    return numerator / denominator


def lattice_distortion(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [lattice_distortion(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    meanR = 0
    for element in alloy.elements:
        meanR += (
            alloy.composition[element]
            * mg.periodic_table.elements[element]["radius"]
        )

    tmp_lattice_distortion = 0
    for i in range(len(alloy.elements) - 1):
        for j in range(i + 1, len(alloy.elements)):
            element = alloy.elements[i]
            otherElement = alloy.elements[j]

            radiusA = (
                alloy.composition[element]
                * mg.periodic_table.elements[element]["radius"]
            )
            radiusB = (
                alloy.composition[element]
                * mg.periodic_table.elements[otherElement]["radius"]
            )

            tmp_lattice_distortion += (
                alloy.composition[element]
                * alloy.composition[otherElement]
                * np.abs(radiusA + radiusB - 2 * meanR)
            ) / (2 * meanR)

    return tmp_lattice_distortion



import inspect

import metallurgy as mg


def is_mod_function(mod, func):
    "checks that func is a function defined in module mod"
    return inspect.isfunction(func) and inspect.getmodule(func) == mod


def list_functions(mod):
    "list of functions defined in module mod"
    return [
        func.__name__
        for func in mod.__dict__.values()
        if is_mod_function(mod, func)
    ]


def calculate(alloys, feature_name):
    if mg.get_model() is not None:
        import cerebral as cb

        model = mg.get_model()
        if feature_name in [
            f["name"] for f in cb.models.get_model_prediction_features(model)
        ]:
            return cb.models.predict(model, alloys)[feature_name]

    if "_linearmix" in feature_name:
        return mg.linear_mixture(alloys, feature_name.split("_linearmix")[0])
    elif "_deviation" in feature_name:
        return mg.deviation(alloys, feature_name.split("_deviation")[0])
    else:
        modules = inspect.getmembers(mg, inspect.ismodule)
        for module in modules:
            for func in list_functions(module[1]):
                if func == feature_name:
                    return getattr(module[1], func)(alloys)

        return mg.linear_mixture(alloys, feature_name)



from __future__ import annotations
import copy
import re
from typing import Union, Callable, Optional
from collections import OrderedDict

import numpy as np
import elementy


class Alloy:
    """An alloy, a mixture of chemical elements with specific percentages.

    Attributes:
        composition: Dictionary matching element symbols to atomic
                            percentages.
        constraints: Dictionary of constraints to follow when adjusting
                            atomic percentages.

    """

    class Composition(dict):
        """Atomic percentages of elements in an alloy.

        Attributes:
            on_change (Callable): Callback function enabling parent
                                  to act on composition changes

        """

        def __init__(self, value: dict, on_change: Callable, *args, **kwargs):
            super().__init__(value, *args, **kwargs)
            self.on_change = on_change

        def __setitem__(self, element: str, percentage: float):
            """
            Set the percentage value of an element in the composition.
            Remove elements from the composition if below minimum threshold.
            """
            if percentage >= 0.01:
                super().__setitem__(element, percentage)
            else:
                if element in self.keys():
                    super().__delitem__(element)
            self.on_change()

    def __init__(
        self,
        composition: Union[str, dict, Alloy],
        constraints: Optional[dict] = None,
        rescale: bool = True,
    ):

        self.composition = parse_composition(composition)

        self.constraints = None

        if rescale:
            self.clamp_composition()
            self.round_composition()

        if constraints is not None:
            self.constraints = parse_constraints(**constraints)
            if rescale:
                self.rescale()

        else:
            self.constraints = None

    def __repr__(self):
        return self.to_string()

    def __eq__(self, other):
        return self.to_string() == other.to_string()

    def __hash__(self):
        return hash(self.to_string())

    @property
    def composition(self) -> Composition:
        """Dictionary of elements and percentages in the alloy."""
        return self._composition

    @composition.setter
    def composition(self, value):
        if isinstance(value, (dict, OrderedDict)):
            value = self.Composition(value, self.on_composition_change)
        self._composition = value

    def on_composition_change(self):
        """Called when composition property changes."""
        self.determine_percentage_constraints()

    @property
    def elements(self) -> list:
        """List of elements in the alloy."""
        return list(self.composition.keys())

    def rescale(self):
        """Adjust elemental percentages to match constraints

        All alloys are constrained such that element percentages sum
        to 1.0. Additional constraints include:

        - Minimum number of unique elements in an alloy.
        - Maximum number of unique elements in an alloy.
        - Maximum and minimum percentage values per element.
        - Precedence order of elements.
        """

        constraints_applied = False
        while not self.constraints_satisfied():

            constraints_applied = True

            precedence = {}
            for element in self.constraints["local_percentages"]:
                if self.constraints["local_percentages"][element]["min"] > 0:
                    if element not in self.composition:
                        self.composition[element] = 0.01

                precedence[element] = self.constraints["local_percentages"][
                    element
                ]["precedence"]

            precedence_order = sorted(
                precedence, key=precedence.get, reverse=True
            )
            reverse_precedence_order = sorted(precedence, key=precedence.get)

            for element in precedence_order:
                if element not in self.composition:
                    if (
                        self.constraints["local_percentages"][element]["min"]
                        > 0
                    ):
                        self.composition[element] = self.constraints[
                            "local_percentages"
                        ][element]["min"]
                    else:
                        continue

                if (
                    self.composition[element]
                    < self.constraints["local_percentages"][element]["min"]
                ):
                    for other_element in reverse_precedence_order:
                        if (
                            element != other_element
                            and other_element in self.composition
                        ):

                            if (
                                self.composition[other_element]
                                > self.constraints["local_percentages"][
                                    other_element
                                ]["max"]
                            ):

                                element_deficit = (
                                    self.constraints["local_percentages"][
                                        element
                                    ]["min"]
                                    - self.composition[element]
                                )

                                other_element_surplus = (
                                    self.composition[other_element]
                                    - self.constraints["local_percentages"][
                                        other_element
                                    ]["max"]
                                )

                                # print("A")
                                # print(self.composition)
                                # print(element, element_deficit)
                                # print(other_element, other_element_surplus)
                                # print()

                                if (
                                    element_deficit >= other_element_surplus
                                    and other_element_surplus > 0
                                ):
                                    self.composition[
                                        element
                                    ] += other_element_surplus
                                    self.composition[
                                        other_element
                                    ] -= other_element_surplus
                                else:
                                    self.composition[
                                        element
                                    ] += element_deficit
                                    self.composition[
                                        other_element
                                    ] -= element_deficit

                                if element not in self.composition:
                                    self.composition[
                                        element
                                    ] = self.constraints["local_percentages"][
                                        element
                                    ][
                                        "min"
                                    ]
                                if (
                                    self.composition[element]
                                    >= self.constraints["local_percentages"][
                                        element
                                    ]["min"]
                                ):
                                    break

                if element not in self.composition:
                    if (
                        self.constraints["local_percentages"][element]["min"]
                        > 0
                    ):
                        self.composition[element] = self.constraints[
                            "local_percentages"
                        ][element]["min"]
                    else:
                        continue

                if (
                    self.composition[element]
                    < self.constraints["local_percentages"][element]["min"]
                ):

                    for other_element in sorted(
                        self.composition.keys(),
                        key=lambda k: np.random.random(),
                    ):

                        if other_element != element:

                            other_element_surplus = self.composition[
                                other_element
                            ]
                            if (
                                other_element
                                in self.constraints["local_percentages"]
                            ):
                                if (
                                    self.constraints["local_percentages"][
                                        element
                                    ]["precedence"]
                                    < self.constraints["local_percentages"][
                                        other_element
                                    ]["precedence"]
                                ):
                                    continue

                                other_element_surplus -= self.constraints[
                                    "local_percentages"
                                ][other_element]["min"]

                            if (
                                normal_round(
                                    other_element_surplus,
                                    self.constraints["sigfigs"] + 1,
                                )
                                > 0
                            ):

                                element_deficit = (
                                    self.constraints["local_percentages"][
                                        element
                                    ]["min"]
                                    - self.composition[element]
                                )

                                # print("B")
                                # print(self.composition)
                                # print(element, element_deficit)
                                # print(other_element, other_element_surplus)
                                # print()

                                if (
                                    element_deficit
                                    >= self.composition[other_element]
                                ):
                                    self.composition[
                                        element
                                    ] += other_element_surplus
                                    self.composition[
                                        other_element
                                    ] -= other_element_surplus
                                else:
                                    self.composition[
                                        element
                                    ] += element_deficit
                                    self.composition[
                                        other_element
                                    ] -= element_deficit

                                if element not in self.composition:
                                    self.composition[
                                        element
                                    ] = self.constraints["local_percentages"][
                                        element
                                    ][
                                        "min"
                                    ]
                                if (
                                    self.composition[element]
                                    >= self.constraints["local_percentages"][
                                        element
                                    ]["min"]
                                ):
                                    break

                if element not in self.composition:
                    if (
                        self.constraints["local_percentages"][element]["min"]
                        > 0
                    ):
                        self.composition[element] = self.constraints[
                            "local_percentages"
                        ][element]["min"]
                    else:
                        continue

                if (
                    self.composition[element]
                    > self.constraints["local_percentages"][element]["max"]
                ):
                    for other_element in reverse_precedence_order:
                        if (
                            element != other_element
                            and other_element in self.composition
                        ):

                            if (
                                self.composition[other_element]
                                < self.constraints["local_percentages"][
                                    other_element
                                ]["min"]
                            ):
                                element_surplus = (
                                    self.composition[element]
                                    - self.constraints["local_percentages"][
                                        element
                                    ]["max"]
                                )

                                other_element_deficit = (
                                    self.constraints["local_percentages"][
                                        other_element
                                    ]["min"]
                                    - self.composition[other_element]
                                )

                                # print("C")
                                # print(self.composition)
                                # print(element, element_surplus)
                                # print(other_element, other_element_deficit)
                                # print()

                                if element_surplus >= other_element_deficit:
                                    self.composition[
                                        element
                                    ] -= other_element_deficit
                                    self.composition[
                                        other_element
                                    ] += element_surplus
                                else:
                                    self.composition[
                                        element
                                    ] -= element_surplus
                                    self.composition[
                                        other_element
                                    ] += element_surplus

                                if element not in self.composition:
                                    self.composition[
                                        element
                                    ] = self.constraints["local_percentages"][
                                        element
                                    ][
                                        "max"
                                    ]
                                if (
                                    self.composition[element]
                                    <= self.constraints["local_percentages"][
                                        element
                                    ]["max"]
                                ):
                                    break

                if element not in self.composition:
                    if (
                        self.constraints["local_percentages"][element]["min"]
                        > 0
                    ):
                        self.composition[element] = self.constraints[
                            "local_percentages"
                        ][element]["min"]
                    else:
                        continue

                if (
                    self.composition[element]
                    > self.constraints["local_percentages"][element]["max"]
                ):
                    for other_element in sorted(
                        self.composition.keys(),
                        key=lambda k: np.random.random(),
                    ):
                        if other_element != element:

                            if (
                                other_element
                                in self.constraints["local_percentages"]
                            ):
                                other_element_deficit = (
                                    self.constraints["local_percentages"][
                                        other_element
                                    ]["max"]
                                    - self.composition[other_element]
                                )
                            else:
                                other_element_deficit = (
                                    1.0 - self.composition[other_element]
                                )

                            element_surplus = (
                                self.composition[element]
                                - self.constraints["local_percentages"][
                                    element
                                ]["max"]
                            )

                            element_surplus = np.max(
                                [
                                    element_surplus,
                                    self.constraints["percentage_step"],
                                ]
                            )
                            other_element_deficit = np.max(
                                [
                                    other_element_deficit,
                                    self.constraints["percentage_step"],
                                ]
                            )

                            # print("D")
                            # print(self.composition)
                            # print(element, element_surplus)
                            # print(other_element, other_element_deficit)
                            # print()

                            if (
                                element_surplus >= other_element_deficit
                                and other_element_deficit > 0
                            ):
                                self.composition[
                                    element
                                ] -= other_element_deficit
                                self.composition[
                                    other_element
                                ] += other_element_deficit
                            else:
                                self.composition[element] -= element_surplus
                                self.composition[
                                    other_element
                                ] += element_surplus

                            if element not in self.composition:
                                self.composition[element] = self.constraints[
                                    "local_percentages"
                                ][element]["max"]
                            if (
                                self.composition[element]
                                <= self.constraints["local_percentages"][
                                    element
                                ]["max"]
                            ):
                                break

            if len(self.composition) > self.constraints["max_elements"]:

                excess = (
                    len(self.composition) - self.constraints["max_elements"]
                )

                toDelete = []
                ordered_elements = sorted(
                    self.composition, key=self.composition.get
                )
                for element in ordered_elements:
                    if element in self.constraints["local_percentages"]:
                        if (
                            self.constraints["local_percentages"][element][
                                "min"
                            ]
                            == 0
                        ):
                            toDelete.append(element)
                    else:
                        toDelete.append(element)

                    if len(toDelete) == excess:
                        break

                for element in toDelete:
                    del self.composition[element]

            elif len(self.composition) < self.constraints["min_elements"]:

                while len(self.composition) < self.constraints["min_elements"]:

                    element_to_add = np.random.choice(
                        self.constraints["allowed_elements"], 1
                    )[0]
                    if element_to_add not in self.composition:
                        self.composition[element_to_add] = np.random.uniform()

            self.clamp_composition()

        if not constraints_applied:
            self.clamp_composition()
        self.round_composition()

    def constraints_satisfied(self) -> bool:
        """Check for satisfaction of the constraints acting on an alloy"""

        satisfied = True
        if self.constraints is not None:

            discrepancy = np.abs(
                1
                - multiple_round(
                    sum(self.composition.values()),
                    self.constraints["percentage_step"],
                )
            )
            if discrepancy > self.constraints["percentage_step"]:
                # print("Sum != 1", sum(self.composition.values()), discrepancy)
                # print(self.composition)
                # print()
                satisfied = False

            if satisfied:
                for element in self.constraints["local_percentages"]:
                    if element in self.composition:
                        if (
                            self.composition[element]
                            < self.constraints["local_percentages"][element][
                                "min"
                            ]
                        ):
                            # print("element below min", element,
                            #       self.composition[element], self.constraints['local_percentages'][element]['min'])
                            # print(self.composition)
                            # print()
                            satisfied = False
                            break
                        elif (
                            self.composition[element]
                            > self.constraints["local_percentages"][element][
                                "max"
                            ]
                        ):
                            # print("element above max", element,
                            #       self.composition[element], self.constraints['local_percentages'][element]['max'])
                            # print(self.composition)
                            # print()
                            satisfied = False
                            break
                    elif (
                        self.constraints["local_percentages"][element]["min"]
                        > 0.0
                    ):
                        # print("element missing", element,
                        #       self.constraints['local_percentages'][element]['min'])
                        # print(self.composition)
                        # print()
                        satisfied = False
                        break

            if satisfied:
                if len(self.composition) > self.constraints["max_elements"]:
                    satisfied = False
                    # print("too many elements")
                    # print(self.composition)
                    # print()
                elif len(self.composition) < self.constraints["min_elements"]:
                    satisfied = False
                    # print("too few elements")
                    # print(self.composition)
                    # print()

        return satisfied

    def clamp_composition(self):
        """Adjust elemental percentages such that they sum to 1.0"""

        total_percentage = sum(self.composition.values())

        for element in self.elements:
            clamped_value = self.composition[element] / total_percentage
            if self.constraints is not None:
                if element in self.constraints["percentages"]:
                    self.composition[element] = max(
                        clamped_value,
                        self.constraints["percentages"][element]["min"],
                    )
                else:
                    self.composition[element] = clamped_value
            else:
                self.composition[element] = clamped_value

    def determine_percentage_constraints(self):
        """Determine the current constraints on an alloy composition.

        When precedence constraints are acting, the maximum and
        minimum percentage values for each element depend on the
        others.
        """

        if self.constraints is None:
            return

        tmp_percentages = copy.deepcopy(self.constraints["percentages"])

        lowest_precedence = {
            "element": None,
            "precedence": np.Inf,
            "percentage": np.Inf,
        }
        for element in self.constraints["percentages"]:
            if element in self.composition:
                if (
                    self.constraints["percentages"][element]["precedence"]
                    <= lowest_precedence["precedence"]
                    and self.constraints["percentages"][element]["precedence"]
                    > 0
                    and self.composition[element]
                    < lowest_precedence["percentage"]
                ):

                    lowest_precedence["precedence"] = self.constraints[
                        "percentages"
                    ][element]["precedence"]
                    lowest_precedence["element"] = element
                    lowest_precedence["percentage"] = self.composition[element]

            if (
                len(
                    self.constraints["percentages"][element][
                        "superior_elements"
                    ]
                )
                > 0
            ):
                superior_elements = []
                for e in self.constraints["percentages"][element][
                    "superior_elements"
                ]:
                    if e in self.composition:
                        superior_elements.append(self.composition[e])
                if len(superior_elements) > 0:
                    tmp_percentages[element]["max"] = min(superior_elements)
                else:
                    tmp_percentages[element]["max"] = 0

        if lowest_precedence["element"] is not None:
            for element in self.composition:
                if element not in tmp_percentages:
                    maxPercent = lowest_precedence["percentage"]

                    tmp_percentages[element] = {
                        "max": maxPercent,
                        "min": 0,
                        "precedence": 0,
                        "superior_elements": [],
                    }

                    for other_element in tmp_percentages:
                        if element != other_element:
                            if (
                                tmp_percentages[element]["precedence"]
                                < tmp_percentages[other_element]["precedence"]
                            ):
                                tmp_percentages[element][
                                    "superior_elements"
                                ].append(other_element)

        self.constraints["local_percentages"] = tmp_percentages

    def to_string(self) -> str:
        """Convert the alloy composition dict to a string"""

        composition_str = ""
        for element in self.elements:
            percentage_str = str(self.composition[element] * 100.0)

            split_str = percentage_str.split(".")
            decimal = split_str[1]
            if decimal == "0":
                percentage_str = split_str[0]
            else:
                decimal_places = len(
                    str(self.composition[element]).split(".")[1]
                )
                percentage_str = str(
                    round(float(percentage_str), decimal_places)
                )

                split_str = percentage_str.split(".")
                decimal = split_str[1]
                if decimal == "0":
                    percentage_str = split_str[0]

            composition_str += element + percentage_str

        return composition_str

    def to_pretty_string(self) -> str:
        """Convert alloy composition to string with LaTeX formatting of subscripts"""
        numbers = re.compile(r"(\d+)")
        return numbers.sub(r"$_{\1}$", self.to_string())

    def round_composition(self):
        """Round elemental percentages in composition while maintaining sum"""

        if len(self.composition) == 0:
            return
        elif len(self.composition) == 1:
            self.composition[self.elements[0]] = 1.0

        if self.constraints is not None:
            sigfigs = self.constraints["sigfigs"]
        else:
            sigfigs = 3

        integer_parts = []
        decimal_parts = []

        elements = list(self.composition.keys())
        for element in elements:

            rounded_percentage = normal_round(
                self.composition[element], sigfigs
            )

            if rounded_percentage > 0:

                percentage = str(self.composition[element] * (10**sigfigs))
                split_percentage = percentage.split(".")

                integer_parts.append(int(split_percentage[0]))

                if len(split_percentage) > 1:
                    decimal_parts.append(
                        float(percentage) - float(split_percentage[0])
                    )
                else:
                    decimal_parts.append(0.0)

        undershoot = (10**sigfigs) - sum(integer_parts)

        if undershoot > 0:
            if self.constraints is not None:

                precedence = {}
                for element in elements:
                    if element in self.constraints["local_percentages"]:
                        precedence[element] = self.constraints[
                            "local_percentages"
                        ][element]["precedence"]
                    else:
                        precedence[element] = 0
                precedence_order = sorted(
                    precedence, key=precedence.get, reverse=True
                )

                filtered_precedence_order = []
                for element in precedence_order:
                    if element in self.composition:
                        filtered_precedence_order.append(element)
            else:
                filtered_precedence_order = list(self.composition.keys())

            i = 0
            while undershoot > 0:
                decimal_part_index = list(self.composition.keys()).index(
                    filtered_precedence_order[i]
                )

                constraints_violated = False
                if self.constraints is not None:
                    if (
                        filtered_precedence_order[i]
                        in self.constraints["local_percentages"]
                    ):
                        if (integer_parts[decimal_part_index] + 1) / (
                            10**sigfigs
                        ) > self.constraints["local_percentages"][
                            filtered_precedence_order[i]
                        ][
                            "max"
                        ]:
                            constraints_violated = True

                if not constraints_violated:
                    integer_parts[decimal_part_index] += 1
                    undershoot -= 1
                i += 1

                if i >= len(filtered_precedence_order):
                    i = 0

            elements = list(self.composition.keys())

            i = 0
            for element in elements:
                self.composition[element] = normal_round(
                    integer_parts[i] / (10**sigfigs), sigfigs
                )
                i += 1

        else:

            elements = list(self.composition.keys())
            i = 0
            for element in elements:

                rounded_value = integer_parts[i] / (10**sigfigs)

                if self.constraints is not None:
                    if element in self.constraints["local_percentages"]:
                        self.composition[element] = max(
                            self.constraints["local_percentages"][element][
                                "min"
                            ],
                            rounded_value,
                        )
                    else:
                        self.composition[element] = rounded_value
                else:
                    self.composition[element] = rounded_value

                i += 1


def parse_composition(composition: Union[str, dict, Alloy]) -> dict:
    """Parse elemental percentages of an alloy from input"""
    if isinstance(composition, str):
        return parse_composition_string(composition)
    elif isinstance(composition, dict):
        return parse_composition_dict(composition)
    elif isinstance(composition, Alloy):
        return composition.composition


def parse_composition_string(composition_string: str) -> dict:
    """Parse elemental percentages of an alloy from a string"""

    composition = {}
    if "(" in composition_string:
        major_composition = composition_string.split(")")[0].split("(")[1]

        major_composition_percentage = (
            float(
                re.split(r"(\d+(?:\.\d+)?)", composition_string.split(")")[1])[
                    1
                ]
            )
            / 100.0
        )

        split_major_composition = re.findall(
            r"[A-Z][^A-Z]*", major_composition
        )

        for element_percentage in split_major_composition:
            split_element_percentage = re.split(
                r"(\d+(?:\.\d+)?)", element_percentage
            )
            composition[split_element_percentage[0]] = (
                float(split_element_percentage[1]) / 100.0
            ) * major_composition_percentage

        minor_composition = composition_string.split(")")[1][
            len(str(int(major_composition_percentage * 100))) :
        ]
        split_minor_composition = re.findall(
            r"[A-Z][^A-Z]*", minor_composition
        )
        for element_percentage in split_minor_composition:
            split_element_percentage = re.split(
                r"(\d+(?:\.\d+)?)", element_percentage
            )

            decimal_places = 2
            if "." in str(split_element_percentage[1]):
                decimal_places += len(
                    str(split_element_percentage[1]).split(".")[1]
                )

            composition[split_element_percentage[0]] = round(
                float(split_element_percentage[1]) / 100.0, decimal_places
            )

    else:

        split_composition = re.findall(r"[A-Z][^A-Z]*", composition_string)

        for element_percentage in split_composition:
            split_element_percentage = re.split(
                r"(\d+(?:\.\d+)?)", element_percentage
            )

            if len(split_element_percentage) > 1:
                decimal_places = 2
                if "." in str(split_element_percentage[1]):
                    decimal_places += len(
                        str(split_element_percentage[1]).split(".")[1]
                    )

                composition[split_element_percentage[0]] = round(
                    float(split_element_percentage[1]) / 100.0, decimal_places
                )
            else:
                composition[split_element_percentage[0]] = 1.0 / len(
                    split_composition
                )

    return filter_order_composition(composition)


def parse_composition_dict(composition: dict) -> dict:
    """Parse elemental percentages of an alloy from a dict"""

    tmp_composition = copy.deepcopy(composition)

    needRescale = False
    for element in tmp_composition:
        if tmp_composition[element] > 1:
            needRescale = True
            break

    if needRescale:
        for element in tmp_composition:
            tmp_composition[element] /= 100.0

    return filter_order_composition(tmp_composition)


def filter_order_composition(composition: dict) -> OrderedDict:
    """Filter & order elements in an alloy composition

    Elements with percentage less than zero are filtered out.
    An OrderedDict is used to sort the composition in descending
    order of elemental percentage.
    """

    filtered_composition = {}
    for element in composition:
        if composition[element] > 0:
            filtered_composition[element] = composition[element]

    ordered_composition = OrderedDict()
    for element in sorted(
        filtered_composition, key=filtered_composition.get, reverse=True
    ):
        ordered_composition[element] = filtered_composition[element]

    return ordered_composition


def parse_constraints(
    min_elements: int = 1,
    max_elements: int = 10,
    percentages: dict = {},
    allowed_elements: list = [e for e in elementy.PeriodicTable().elements],
    disallowed_elements: list = [],
    sigfigs: int = 3,
    percentage_step: float = 0.01,
) -> dict:
    """Parse constraint rules from input

    Args:
        min_elements: Minimum number of elements in an alloy
        max_elements: Maximum number of elements in an alloy
        percentages: Dict of maximum and minimum percentages
                            per element
        allowed_elements: List of elements allowed in an alloy
        disallowed_elements: List of elements not allowed in an
                             alloy
        sigfigs: Number of significant figures to consider for
                       percentages
        percentage_step: Increment between percentages
    """

    if not isinstance(percentages, dict):
        if isinstance(percentages, list):
            tmp_percentages = {}
            for element in percentages:
                tmp_percentages[element] = {}
            percentages = tmp_percentages

    for element in percentages:
        if "min" not in percentages[element]:
            percentages[element]["min"] = 0.0
        else:
            percentages[element]["min"] = max(percentages[element]["min"], 0.0)

        if "max" not in percentages[element]:
            percentages[element]["max"] = 1.0
        else:
            percentages[element]["max"] = min(percentages[element]["max"], 1.0)

        if "precedence" not in percentages[element]:
            percentages[element]["precedence"] = 0

    for element in percentages:
        percentages[element]["superior_elements"] = []
        for other_element in percentages:
            if element != other_element:
                if (
                    percentages[element]["precedence"]
                    < percentages[other_element]["precedence"]
                ):
                    percentages[element]["superior_elements"].append(
                        other_element
                    )

    min_sum = sum([percentages[e]["min"] for e in percentages])
    if min_sum > 1:
        print(
            "Impossible constraints, mins for each element sum greater than 1"
        )

    for element in disallowed_elements:
        if element in allowed_elements:
            allowed_elements.remove(element)

    return {
        "percentages": percentages,
        "local_percentages": percentages,
        "allowed_elements": allowed_elements,
        "min_elements": min_elements,
        "max_elements": max_elements,
        "sigfigs": sigfigs,
        "percentage_step": percentage_step,
    }


def normal_round(num: float, ndigits: int = 0) -> float:
    """Rounds a float to the specified number of decimal places.

    Args:
        num: the value to round
        ndigits: the number of digits to round to
    """
    if ndigits == 0:
        return int(num + 0.5)
    else:
        digit_value = 10**ndigits
        return int(num * digit_value + 0.5) / digit_value


def multiple_round(
    num: float, multiple: Union[float, int]
) -> Union[float, int]:
    """Round a number to the nearest multiple of another number.

    Args:
        num: Value to round
        multiple: Multiple-of to round towards
    """
    return multiple * round(num / multiple)



from collections.abc import Iterable

import metallurgy as mg
from . import linear_mixture
from .alloy import Alloy


def price(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [price(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    total_weight = linear_mixture(alloy, "mass")

    total_price = 0
    for element in alloy.elements:
        if mg.periodic_table.elements[element]["price"] is not None:
            weight_percent = (
                mg.periodic_table.elements[element]["mass"]
                * alloy.composition[element]
                / total_weight
            )
            total_price += (
                weight_percent * mg.periodic_table.elements[element]["price"]
            )
        else:
            return None

    return total_price



import pytest

import metallurgy as mg


def test_random_alloy(random_alloy=None):
    if random_alloy is None:
        random_alloy = mg.generate.random_alloy()
    assert sum(random_alloy.composition.values()) == pytest.approx(1.0)


def test_random_alloys():

    num_alloys = 100
    random_alloys = mg.generate.random_alloys(num_alloys)

    assert len(random_alloys) == num_alloys

    for random_alloy in random_alloys:
        test_random_alloy(random_alloy)


def test_generate_binary_alloys():

    elements = ["Cu", "Zr"]

    step = 1
    alloys, percentages = mg.generate.binary(elements, step)
    assert len(alloys) == len(percentages) == 101
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[1]: 1.0}

    step = 2
    alloys, percentages = mg.generate.binary(elements, step)
    assert len(alloys) == len(percentages) == 51
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[1]: 1.0}

    step = 0.5
    alloys, percentages = mg.generate.binary(elements, step)
    assert len(alloys) == len(percentages) == 201
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[1]: 1.0}


def test_generate_ternary_alloys():

    elements = ["Cu", "Zr", "Al"]

    step = 1
    alloys, percentages = mg.generate.ternary(elements, step)
    assert len(alloys) == len(percentages) == 5151
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[2]: 1.0}

    step = 2
    alloys, percentages = mg.generate.ternary(elements, step)
    assert len(alloys) == len(percentages) == 1326
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[2]: 1.0}


def test_requirements():

    num_alloys = 100

    percentage_constraints = {
        "Cu": {"min": 0.2, "max": 0.6},
        "Fe": {"min": 0.0, "max": 1.0},
        "Ni": {"min": 0.01, "max": 0.99},
    }
    random_alloys = mg.generate.random_alloys(
        num_alloys, percentage_constraints=percentage_constraints
    )

    for random_alloy in random_alloys:
        for element in percentage_constraints:
            if percentage_constraints[element]["min"] > 0:
                assert element in random_alloy.composition

            if element in random_alloy.composition:
                assert (
                    random_alloy.composition[element]
                    >= percentage_constraints[element]["min"]
                )
                assert (
                    random_alloy.composition[element]
                    <= percentage_constraints[element]["max"]
                )


def test_mixture():

    A = mg.Alloy("Cu50Zr50")

    mixed = mg.generate.mixture([A])

    assert mixed.composition == A.composition

    B = mg.Alloy("Fe50Ni50")

    mixed = mg.generate.mixture([A, B])

    assert mixed.composition == {
        "Cu": 0.25,
        "Zr": 0.25,
        "Fe": 0.25,
        "Ni": 0.25,
    }

    mixed = mg.generate.mixture([A, B], [0.9, 0.1])

    assert mixed.composition == {
        "Cu": 0.45,
        "Zr": 0.45,
        "Fe": 0.05,
        "Ni": 0.05,
    }

    C = mg.Alloy(
        "Cu50Zr50",
        constraints={
            "percentages": {"Cu": {"min": 0.1, "max": 0.6}},
            "max_elements": 2,
            "min_elements": 1,
        },
    )

    mixed = mg.generate.mixture([C])

    assert mixed.composition == C.composition

    mixed = mg.generate.mixture([C, B])

    assert len(mixed.elements) <= 2
    assert mixed.composition["Cu"] <= 0.6
    assert mixed.composition["Cu"] >= 0.1



from collections.abc import Iterable

import numpy as np

import metallurgy as mg
from .alloy import Alloy


def ideal_entropy(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [ideal_entropy(a) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    total_ideal_entropy = 0
    for element in alloy.elements:
        total_ideal_entropy += alloy.composition[element] * np.log(
            alloy.composition[element]
        )

    return -total_ideal_entropy


def ideal_entropy_xia(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [ideal_entropy_xia(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    cube_sum = 0
    for element in alloy.elements:
        cube_sum += (
            alloy.composition[element]
            * mg.periodic_table.elements[element]["atomic_volume"]
        )

    ideal_entropy = 0
    for element in alloy.composition:
        ideal_entropy += alloy.composition[element] * np.log(
            (
                alloy.composition[element]
                * mg.periodic_table.elements[element]["atomic_volume"]
            )
            / cube_sum
        )

    return -ideal_entropy


def mismatch_entropy(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mismatch_entropy(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    if len(alloy.elements) == 1:
        return 0.0

    diameters = {}
    for element in alloy.composition:
        if mg.periodic_table.elements[element]["radius"] is None:
            return None

        diameters[element] = mg.periodic_table.elements[element]["radius"] * 2

    sigma_2 = 0
    for element in alloy.composition:
        sigma_2 += alloy.composition[element] * (diameters[element] ** 2)

    sigma_3 = 0
    for element in alloy.composition:
        sigma_3 += alloy.composition[element] * (diameters[element] ** 3)

    y_3 = (sigma_2**3) / (sigma_3**2)

    y_1 = 0
    y_2 = 0

    for i in range(len(alloy.elements) - 1):
        for j in range(i + 1, len(alloy.elements)):
            element = alloy.elements[i]
            otherElement = alloy.elements[j]

            y_1 += (
                (diameters[element] + diameters[otherElement])
                * ((diameters[element] - diameters[otherElement]) ** 2)
                * alloy.composition[element]
                * alloy.composition[otherElement]
            )

            y_2 += (
                diameters[element]
                * diameters[otherElement]
                * ((diameters[element] - diameters[otherElement]) ** 2)
                * alloy.composition[element]
                * alloy.composition[otherElement]
            )

    y_1 /= sigma_3

    y_2 *= sigma_2 / (sigma_3**2)

    packing_fraction = 0.64
    zeta = 1.0 / (1 - packing_fraction)

    return (
        ((3.0 / 2.0) * ((zeta**2) - 1) * y_1)
        + ((3.0 / 2.0) * ((zeta - 1) ** 2) * y_2)
        - (1 - y_3) * (0.5 * (zeta - 1) * (zeta - 3) + np.log(zeta))
    )


def mixing_entropy(alloy):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_entropy(a) for a in alloy]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    ideal = ideal_entropy(alloy)
    mismatch = mismatch_entropy(alloy)

    if ideal is not None and mismatch is not None:
        return ideal + mismatch
    else:
        return None



units = {
    "Dmax": "mm",
    "Tl": "K",
    "Tg": "K",
    "Tx": "K",
    "deltaT": "K",
    "price_linearmix": "\\$/kg",
    "price": "\\$/kg",
    "mixing_enthalpy": "kJ/mol",
    "mixing_Gibbs_free_energy": "kJ/mol",
}
inverse_units = {}
for feature in units:
    if "/" not in units[feature]:
        inverse_units[feature] = "1/" + units[feature]
    else:
        split_units = units[feature].split("/")
        inverse_units[feature] = split_units[1] + "/" + split_units[0]


def pretty_name(feature_name):

    name = ""
    feature_parts = feature_name.split("_")
    if "linearmix" in feature_name or "deviation" in feature_name:
        if len(feature_parts) > 1:
            if feature_parts[-1] == "linearmix":
                name = r"$\Sigma$ "
            elif feature_parts[-1] == "deviation":
                name = r"$\delta$ "
            name += " ".join(word.title() for word in feature_parts[0:-1])
    else:
        name += " ".join(word.title() for word in feature_parts)
    return name



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.collections as mcoll
import matplotlib as mpl
import ternary as ternary_plt

plt.style.use("ggplot")
plt.rc("axes", axisbelow=True)


def binary(
    alloys,
    data,
    xlabel=None,
    ylabel=None,
    labels=None,
    scatter_data=None,
    use_colorline=False,
    save_path=None,
):
    fig = plt.figure()
    ax1 = fig.add_subplot(111)

    elements = []
    for alloy in alloys:
        for element in alloy.elements:
            if element not in elements:
                elements.append(element)

    percentages = []
    for alloy in alloys:
        if elements[0] in alloy.composition:
            percentages.append(alloy.composition[elements[0]] * 100)
        else:
            percentages.append(0)

    percentages, data = zip(*sorted(zip(percentages, data)))

    if use_colorline:
        lc = colorline(data[0], data[1], ax1, z=percentages)
        cbar = plt.colorbar(lc)
        cbar.set_label(elements[0] + " %", rotation=270)

    elif isinstance(data[0], list):
        for i in range(len(data)):
            label = None
            if labels is not None:
                label = labels[i]

            ax1.plot(percentages, data[i], label=label)
    else:
        ax1.plot(percentages, data)

    if scatter_data is not None:
        for scatter_datum in scatter_data:
            ax1.scatter(
                scatter_datum["data"][0],
                scatter_datum["data"][1],
                marker=scatter_datum["marker"],
                label=scatter_datum["label"],
                edgecolors="k",
                zorder=20,
            )
        ax1.legend(loc="best")
    ax1.autoscale()

    if not isinstance(data[0], list) and not use_colorline:
        ax2 = ax1.twiny()
        ax2.set_xticks(ax1.get_xticks())
        ax2.set_xbound(ax1.get_xbound())
        ax2.set_xticklabels([int(100 - x) for x in ax1.get_xticks()])

        ax1.set_xlabel(elements[0] + " %")
        ax2.set_xlabel(elements[1] + " %")
        ax2.grid(False)

    if xlabel is not None:
        ax1.set_xlabel(xlabel)
    if ylabel is not None:
        ax1.set_ylabel(ylabel)

    plt.tight_layout()

    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close()


def ternary_heatmap(
    alloys,
    data,
    step=0.01,
    scatter_data=None,
    save_path=None,
    label=None,
    title=None,
    vmin=None,
    vmax=None,
    ax=None,
    showColorbar=True,
):

    scale = 1 / step
    multiple = 0.1 / step
    fontsize = 10
    tick_fontsize = 6
    tick_offset = 0.018
    gridline_width = 0.3
    gridline_style = "--"
    gridline_color = "white"

    elements = []
    for alloy in alloys:
        for element in alloy.elements:
            if element not in elements:
                elements.append(element)

    if not isinstance(data, dict):
        data = list(data)

        heatmap_data = dict()

        for i in range(len(data)):
            percent_A = 0
            percent_B = 0

            if elements[0] in alloys[i].composition:
                percent_A = round(alloys[i].composition[elements[0]] / step, 2)
            if elements[1] in alloys[i].composition:
                percent_B = round(alloys[i].composition[elements[1]] / step, 2)

            heatmap_data[(percent_A, percent_B)] = data[i]

        data = heatmap_data

    if ax is None:
        figure, tax = ternary_plt.figure(scale=scale)
    else:
        figure, tax = ternary_plt.figure(scale=scale, ax=ax)

    tax.get_axes().axis("off")

    tax.gridlines(
        color=gridline_color,
        multiple=multiple,
        linewidth=gridline_width,
        ls=gridline_style,
    )

    tax.set_axis_limits({"b": [0, 100], "l": [0, 100], "r": [0, 100]})
    tax.get_ticks_from_axis_limits(multiple=multiple)
    tax.set_custom_ticks(
        fontsize=tick_fontsize, offset=tick_offset, multiple=multiple
    )

    tax.left_axis_label(elements[2] + " %", fontsize=fontsize, offset=0.12)
    tax.right_axis_label(elements[1] + " %", fontsize=fontsize, offset=0.12)
    tax.bottom_axis_label(elements[0] + " %", fontsize=fontsize, offset=0.12)
    tax.clear_matplotlib_ticks()

    tax.set_title(title, pad=15)

    jet_cmap = mpl.cm.get_cmap("jet")
    tax.heatmap(
        heatmap_data,
        cmap=jet_cmap,
        vmax=vmax,
        vmin=vmin,
        cbarlabel=label,
        colorbar=showColorbar,
    )

    if scatter_data is not None:
        for scatter_datum in scatter_data:
            tax.scatter(
                scatter_datum["data"],
                marker=scatter_datum["marker"],
                label=scatter_datum["label"],
                edgecolors="k",
                zorder=20,
            )

        tax.legend(loc="upper right", handletextpad=0.1, frameon=False)

    tax.get_axes().set_aspect(1)
    tax._redraw_labels()

    if ax is None:
        if save_path is None:
            tax.show()
        else:
            tax.savefig(save_path + ".png")

        tax.close()
        figure.clf()


def colorline(
    x,
    y,
    ax,
    z=None,
    cmap="jet",
    norm=plt.Normalize(0.0, 100.0),
    linewidth=2,
    alpha=1.0,
):
    """
    http://nbviewer.ipython.org/github/dpsanders/matplotlib-examples/blob/master/colorline.ipynb
    http://matplotlib.org/examples/pylab_examples/multicolored_line.html
    Plot a colored line with coordinates x and y
    Optionally specify colors in the array z
    Optionally specify a colormap, a norm function and a line width
    """

    # Default colors equally spaced on [0,1]:
    if z is None:
        z = np.linspace(0.0, 1.0, len(x))

    # Special case if a single number:
    # to check for numerical input -- this is a hack
    if not hasattr(z, "__iter__"):
        z = np.array([z])

    z = np.asarray(z)

    segments = make_segments(x, y)

    lc = mcoll.LineCollection(
        segments,
        array=z,
        cmap=cmap,
        norm=norm,
        linewidth=linewidth,
        alpha=alpha,
    )

    ax.add_collection(lc)
    return lc


def make_segments(x, y):
    """
    Create list of line segments from x and y coordinates, in the correct format
    for LineCollection: an array of the form numlines x (points per line) x 2 (x
    and y) array
    """

    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    return segments



import numbers
from collections.abc import Iterable

import metallurgy as mg
from .alloy import Alloy


def linear_mixture(alloy, feature_name):
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [linear_mixture(a, feature_name) for a in list(alloy)]
    elif not isinstance(alloy, Alloy):
        alloy = Alloy(alloy)

    mixed_property = 0
    for element in alloy.elements:
        value = mg.periodic_table.elements[element][feature_name]
        if value is None:
            return None
        elif isinstance(value, list):
            value = value[0]

        if not isinstance(value, numbers.Number):
            return None

        mixed_property += alloy.composition[element] * value

    return mixed_property
